###############################################################################
# gpUtils 1.0
#
# General Purpose functions for snackAmp Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
###############################################################################
set [file root [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/gpUtils.tcl,v 1.37 2003/05/17 14:26:52 wilkason Exp $
}
###############################################################################
# Function   : K
# Description: K Combinator to work around tcl internal copy on lrepace/linsert
# Author     : Tom Wilkason
###############################################################################
proc K {a b} {set a}
###############################################################################
# Function   : ldelete
# Description: Remove an entry from a list by index
# Author     : Tom Wilkason
###############################################################################
proc ldelete {listName index} {
    upvar $listName list
    set list [lreplace [K $list [set list {}]] $index $index]
}
###############################################################################
# Function   : lset
# Description: Replace an entry in a list by index
#              Note: part of core with 8.4+
# Author     : Tom Wilkason
###############################################################################
if {[info command lset] == {}} {
   proc lset {listName index val} {
       upvar $listName list
       set list [lreplace [K $list [set list {}]] $index $index $val]
   }
} ;# end if
###############################################################################
# Function   : shuffle
# Description: Shuffle a list, used to randomize play list
# Author     : Tom Wilkason
###############################################################################
proc shuffle {list} {
   set tmp [list]
   foreach item $list {
      lappend tmp [list $item [expr {rand()}]]
   }
   set res [list]
   foreach item [lsort -real -index 1 $tmp] {
      foreach {orig num} $item {
         lappend res $orig
      } ;# end foreach
   }
   return $res
}
###############################################################################
# Function   : shuffleByFolder
# Description: Shuffle a list by folder, keep folder contents sorted
# Author     : Tom Wilkason
###############################################################################
proc shuffleByFolder {list {cmd lsort}} {
   set tmp [list]
   ;# Create array of folders containing the files within
   foreach item $list {
      lappend fcontents([file dirname $item]) $item
   }
   ;# Shuffle the folder names
   foreach item [array names fcontents] {
      lappend tmp [list $item [expr {rand()}]]
   }
   set res [list]
   ;# For each folder, append the contents sorted
   foreach item [lsort -real -index 1 $tmp] {
      foreach {orig num} $item {
         foreach {file} [$cmd $fcontents($orig)] {
            lappend res $file
         } ;# end foreach
      } ;# end foreach
   }
   return $res
}
###############################################################################
# Function   : ldsort
# Description: Lsort by dictionary
# Author     : Tom Wilkason
# Date       : 12/28/2002
###############################################################################
proc ldsort {list} {
   return [lsort -dictionary $list]
}
###############################################################################
# Function   : commify
# Description: Insert commas into an integer for display
# Author     : Jeff Hobbs
###############################################################################
proc commify {num {sep ,}} {
    while {[regsub {^([-+]?\d+)(\d\d\d)} $num "\\1$sep\\2" num]} {}
    return $num
}
###############################################################################
# Function   : timify
# Description: Format a clock seconds into a formatted duration
# Author     : Jeff Hobbs
###############################################################################
proc timify {len} {
   if {[catch {clock format [expr {int($len)}] -format "%T" -gmt 1} res]} then {
      set res "00:00:00"
   }
   return $res
}
###############################################################################
# Function   : stringTitle
# Description: Routine to properly case a string, better than string totitle
# Author     : Tom Wilkason
###############################################################################
proc stringTitle {String} {
   set isWordStart 1
   set result ""
   foreach {c} [split $String {}] {
      ;# Allow possesives
      set isAlpha [expr {[string is wordchar $c] | $c == "'"}]
      if {$isWordStart} {
         if {$isAlpha} {
            append result [string toupper $c]
            set isWordStart 0
         } else {
            append result $c
         }
      } else {
         if {$isAlpha} {
            append result [string tolower $c]
         } else {
            set isWordStart 1
            append result $c
         }
      }
   } ;# end foreach
   return $result
}
###############################################################################
# Function   : chooseFile
# Description: Bring up file selector to select a file. Assign it to the var
# Author     : Laurent Duperval
# Date       : 1/12/2002
###############################################################################
proc chooseFile {{start .} {parent .}} {
   if {$start == "."} {
      set start [pwd]
   } ;# end if
   set choice [tk_getOpenFile -initialdir $start -parent $parent]
   if {$choice == {}} {
      return {}
   } else {
      return [casedDirName $choice 1]
   }
}
###############################################################################
# Function   : cleanPathList
# Description: Correct normal goofs in a path name and clean it up
#              so it looks like a unix path name. Doesn't verify the
#              file actually exists.
# Author     : Tom Wilkason
###############################################################################
proc cleanPathList {_Var {check 0}} {
   upvar $_Var Path
   if {$Path != {}} {
      set result [list]
      foreach {token} $Path {
         if {$token != {}} {
            lappend result [casedDirName $token $check]
         } ;# end if
      } ;# end foreach
      set Path [string map {\{ \" \} \"} $result]      ;#"
   }
}
###############################################################################
# Function   : casedDirName
# Description: Correct the file case to match what is actually on the hard
#              disk rather than what was input. Windows only issue.
#              Note: Will not work with UNC because glob wont
# Author     : Tom Wilkason
# Date       : 3/12/2001
###############################################################################
proc casedDirName {choice {verify 0}} {
   if {$choice != {}} {
      set choice [eval file join [file split [string trim $choice]]]
      if {$verify && ($::tcl_platform(platform) == "windows")} {
         while {1} {
            if {[catch {set choice [file attributes $choice -longname]} message]} {
               set answer [getEntryDialog "'$choice' cannot be found, please correct it or 'cancel'" choice]
               if {$answer == "cancel"} {
                  break
               } ;# end if
            } else {
               break
            }
         }
      }
   }
   return $choice
}
###############################################################################
# Function   : chooseDirectory
# Description: Wrapper for tk_chooseDirectory so it doesn't return
#              uppercase names on Windows (yuck!)
# Author     : Tom Wilkason
###############################################################################
proc chooseDirectory {{start .} {parent .} {title "Please Choose a Directory"}} {
   if {$start == "."} {
      set start [pwd]
   } ;# end if
   set choice [tk_chooseDirectory -initialdir $start -parent $parent -title $title]
   if {$choice == {}} {
      return {}
   } else {
      return [casedDirName $choice 1]
   }
}
###############################################################################
# Function   : cleanKeyName
# Description: Clean up a tk key name to a shortened version
###############################################################################
proc cleanKeyName {name} {
   set name [string map {< "" > "" Control Ctrl Key- ""} $name]
}
###############################################################################
# Function   : browser namespace
# Description: Used to invoke the http browser on the help docs
###############################################################################
namespace eval browser {
   variable playControl
   variable AMGSearch
   variable allmusicBase
   variable cddbBase
   array set AMGSearch {
      Artist 1
      Album 2
      Title 6
      Genre 4
   }
   set allmusicBase {http://allmusic.com/cg/amg.dll?p=amg&sql=} ;#1Eva|Cassidy
   set cddbBase {http://www.freedb.org/freedb_search.php?words=} ;#Eva+Cassidy&allfields=NO&fields=artist&fields=title
   #  cddbBase {http://www.Gracenote.com/xm/search?q=} ;#1Eva+Cassidy&f=artist
}
###############################################################################
# Function   : browser::help
# Description: Launch help using the browser
# Author     : Tom Wilkason
# Date       : 12/30/2001
###############################################################################
proc browser::init {args} {
   ;##
   ;# Now check real folders
   ;#
   variable helpFile
   if {![info exists helpFile]} {
      foreach {location} {user_manual/index.html} {
         set name [file join $::TA(webRoot) $location]
         if {[file exists $name]} {
            set helpFile $location
            puts "Found help file $location"
            break
         } else {
            puts "cant find $name"
         }
      } ;# end foreach
   }
}
proc browser::help {{tag {}}} {
   variable helpFile
   if {[info exists helpFile]} {
      if {$tag=={}} {
         urlOpen $helpFile
      } else {
         urlOpen $helpFile#$tag
      }
   } else {
      tk_messageBox -type ok -icon warning \
         -message "No Help File available or web server is disabled"
   }

}
###############################################################################
# Function   : browser::findExecutable
# Description: Pulled from the wiki http://mini.net/cgi-bin/wikit/557.html
###############################################################################
proc browser::findExecutable {progname varname} {
   set progs [auto_execok $progname]
   if {[llength $progs]} {
       set pattern {([^[:space:]]+)\((.*)\)}
       set isArray [regexp $pattern $varname {} arr index]
       if {$isArray == 0} {
           upvar 1 $varname result
           set result [lindex $progs 0]
       } else {
           upvar $arr array
           set array($index) [lindex $progs 0]
       }
   }
   return [llength $progs]
}

###############################################################################
# Function   : browser::allMusic
# Description: Launch an allmusic query on passed in data
###############################################################################
proc browser::allMusic {_search {type Artist}} {
   variable AMGSearch
   variable allmusicBase
   upvar $_search search
   set redir {<meta http-equiv="refresh" content="0; URL=$url">}
   set parts [join [split $search] |]
   if {[catch {set url "$allmusicBase$AMGSearch($type)$parts"} result]} {
      return
   }

   set redir [subst $redir]
   set redirFile [file join $::HOME redirect.html]
   if {[catch {open $redirFile w} result]} {
      saLog "Error:$result"
      return
   } else {
      puts $result $redir
      close $result
   }
   #puts $url
   #puts $redirFile

   urlOpen $redirFile
}
###############################################################################
# Function   : browser::allMusic
# Description: Launch an cddb query on passed in data
###############################################################################
proc browser::cddb {_Array} {
   package require HttpD
   upvar $_Array Array
   variable cddbBase
   set redir {<meta http-equiv="refresh" content="0; URL=$url">}
   set url $cddbBase
   set Tokens [list]
   set Fields [list]
   foreach {field} [array names Array] {
      if {$field == "Artist" || $field == "Title"} {
         ;# Improve on the cleaning up, use real http substituion
         set contents [string trim [HttpD::Encode $Array($field)]]
         if {[string length $contents] > 0} {
            lappend Fields "fields=[string tolower $field]"
            eval lappend Tokens $contents
         } ;# end if
      } ;# end if
   } ;# end foreach
   append url [join $Tokens +]
   append url "&allfields=NO&allcats=YES&grouping=none&"
   append url [join $Fields &]
   ;# Create the redirection file
   set redir [subst $redir]
   set redirFile [file join $::HOME redirect.html]
   if {[catch {open $redirFile w} result]} {
      saLog "Error:$result"
      return
   } else {
      puts $result $redir
      close $result
   }
   urlOpen $redirFile
}
###############################################################################
# Function   : browser::urlOpen
# Description: Pulled from the wiki http://mini.net/cgi-bin/wikit/557.html
###############################################################################
proc browser::urlOpen {url} {
   global env tcl_platform
   switch $tcl_platform(platform) {
      "unix" {
         expr {
            [info exists env(BROWSER)] ||
            [findExecutable netscape       env(BROWSER)] ||
            [findExecutable iexplorer      env(BROWSER)] ||
            [findExecutable $env(NETSCAPE) env(BROWSER)] ||
            [findExecutable lynx           env(BROWSER)]
         }
         ;# Support files from the ~ folder
         if {[file exists $url]} {
            set url [file nativename $url]
         }
         # lynx can also output formatted text to a variable
         # with the -dump option, as a last resort:
         # set formatted_text [ exec lynx -dump $url ] - PSE
         if {[catch {exec $env(BROWSER) $url > /dev/null < /dev/null &}]} {
            if {[catch {exec $env(BROWSER) -remote openURL\(file:$url\) > /dev/null < /dev/null  &}]} {
               # perhaps browser doesn't understand -remote flag
               if {[catch {exec $env(BROWSER) $url > /dev/null < /dev/null &} emsg]} {
                  saLog "Error displaying $url in browser '$env(BROWSER)'\n$emsg"
               }
            }
         }
      }
      "windows" {
         if {$tcl_platform(os) == "Windows NT"} {
            if {[catch {file attributes $url -shortname} surl]} then {
               set rc [catch {exec $env(COMSPEC) /c start $url &} emsg]
               #puts "No local file"
            } else {
               set rc [catch {exec $env(COMSPEC) /c start $surl &} emsg]
               #puts "Using local file"
            }
         } else {
            # Windows 95/98
            if {[catch {file attributes $url -shortname} surl]} then {
               set rc [catch {exec start $url &} emsg]
            } else {
               set rc [catch {exec start $surl &} emsg]
            }
         }
         if {$rc} {
            saLog "Error displaying $url in browser\n$emsg"
         }
      }
      "macintosh" {
         if {0 == [info exists env(BROWSER)]} {
            set env(BROWSER) "Browse the Internet"
         }
         if {[catch {
               AppleScript execute\
                     "tell application \"$env(BROWSER)\"
               open url \"$url\"
               end tell
               "} emsg]
         } then {
            saLog "Error displaying $url in browser\n$emsg"
         }
      }
   } ;## end of switch
}
###############################################################################
# Function   : hasData
# Description: Return true if a array field has data
# Author     : Tom Wilkason
# Date       : 2/7/2001
###############################################################################
proc hasData {_Data element} {
   upvar $_Data Data
   return [expr {[info exists Data($element)] && [string length $Data($element)] > 0}]
}
###############################################################################
# Function   : debug
# Description: Debug output
###############################################################################
proc debug {Message} {
   set Stack [list]
   set depth [info level]
   set lastArgs [list]
   for {set i 1} {$i < $depth} {incr i} {
      lappend Stack [lindex [info level $i] 0]
      set lastArgs [lrange [info level $i] 1 end]
   } ;# end for
   saLog "Debug:[join $Stack =>] \{$lastArgs\} : $Message"
}
;##
;# Utility function to show call stack frame
;#
proc Trace {args} {
   if {$::snackAmpSettings(Trace)} {
      set Stack [list]
      set depth [info level]
      set lastArgs [list]
      for {set i 1} {$i < $depth} {incr i} {
         lappend Stack [lindex [info level $i] 0]
         set lastArgs [lrange [info level $i] 1 end]
      } ;# end for
      saLog "[join $Stack =>] \{$lastArgs\}"
   } ;# end if
}
##
# Return proc name of calling function
#
proc me {args} {
   return [lindex [info level 1] 0]
}

;##
;# Namespace to watch textvariables linked to an entry box and
;# color them based on the data changed
;# Binding for ctrl-z to undo changes
namespace eval widgetWatch {
   variable Widgets
}
;##
;# User callable routine to set a watch on an
;# entry widget text variable, sets the
;# text background pink when the data changes
;#
proc widgetWatch::setWatch {widget _Element} {
   if {[catch {
      upvar #0 $_Element element
      set cleanElement [string trimleft $_Element :]
      variable Widgets
      set Widgets($widget) $widget
      set Widgets($widget,Orig) $element
      set Widgets($widget,Redo) $element
      set Widgets($widget,bg) [$widget cget -background]
      trace variable $_Element w [list widgetWatch::setCallback $widget]
      trace variable $_Element u [list widgetWatch::resetCallback $widget]
      ;# Undo/Redo keys
      bind $widget <Control-Key-z> [list widgetWatch::undo $widget $_Element]
      bind $widget <Control-Key-y> [list widgetWatch::redo $widget $_Element]
   } result] } then {
      puts "setWatch $result\n$::errorInfo"
   }
}
;##
;# Undo a setting back to the original value
;#
proc widgetWatch::undo {widget _Glob} {
   variable Widgets
   set Widgets($widget,Redo) [set $_Glob]
   set $_Glob $Widgets($widget,Orig)
}
;##
;# Redoe a setting just prior to the last undo
;#
proc widgetWatch::redo {widget _Glob} {
   variable Widgets
   set $_Glob $Widgets($widget,Redo)
}
;##
;# Define the callback to trap modifications to variables
;#
proc widgetWatch::setCallback {widget _Glob item args} {
   variable Widgets
   if {[catch {
      upvar $_Glob GlobArray
      set value $GlobArray($item)
      set ovalue $Widgets($widget,Orig)
      #puts "'$value' vs '$ovalue'"
      if {[string equal $value $ovalue]} {
         #puts "Same for $Widgets($widget) $Widgets($widget,bg)"
         $widget configure -background $Widgets($widget,bg)
      } else {
         #puts "Different for $Widgets($widget)"
         $widget configure -background pink
      }
   } result] } then {
      puts "setCallback $result\n$::errorInfo"
   }
}
;##
;# Callback when the variable is unset, this
;# removes local copies data
;#
proc widgetWatch::resetCallback {widget _Glob item args} {
   variable Widgets
   if {[catch {
      unset Widgets($widget)
      unset Widgets($widget,bg)
      unset Widgets($widget,Orig)
      unset Widgets($widget,Redo)
   } result] } then {
      puts "resetCallback $result\n$::errorInfo"
   }
}
;##
;# Hand type in to get memory usage of globals
;#
proc usage {args} {
   #foreach ns [namespace children ::] {
   #} ;# end foreach
   foreach var [lsort -dictionary [info globals]] {
      upvar $var Var
      if {[array exists ::$var]} {
         set names [llength [array names Var]]
         puts "Array   : $var has [string length [array get Var]] bytes, $names elements,"
      } else {
         puts "Variable: $var has [string length $Var] bytes"
      }
   }
}
###############################################################################
# Function   : safeWrite
# Description: Write a file out safely
# Author     : Tom Wilkason
# Date       : 5/5/2003
###############################################################################
proc safeWrite {file contents} {
   ;# Write to temp file
   set tfile [file join [file dirname $file] [clock clicks].dat]
   if {[catch {open $tfile w} OutFile]} {
      return -code error "Save Error: Unable to open '$file' for writing\n$OutFile"
   }
   if {[catch {
      fconfigure $OutFile -buffersize 100000
      puts -nonewline $OutFile $contents
      close $OutFile
   } result] } then {
      catch {close $OutFile}
      return -code error $result
   }

   ;# Swap Names
   if {[catch {
      ;# Now swap names
      if {[file exists $file]} {
         file delete -force $file
      }
      file rename $tfile $file
   } result] } then {
      file delete -force $tfile
      return -code error $result
   } else {
      return 1
   }
}
;##
;# Some tools to deal with ip address, can use to mask incoming connections
;#
namespace eval iptools {
   proc myIP {args} {
   	# find out localhost's IP address
   	# courtesy David Gravereaux, Heribert Dahms
   	set TheServer [socket -server none -myaddr [info hostname] 0]
   	set MyIP [lindex [fconfigure $TheServer -sockname] 0]
   	close $TheServer
   	return $MyIP
   }
   proc subNetMask {ip} {
      foreach {a b c d} [split $ip .] {break}
      if {$a < 128} {
        # class A
        return 255.0.0.0
      } elseif {$a < 192} {
        # class B
        return 255.255.0.0
      } else {
        # class C
        return 255.255.255.0
      }
   }
   proc meetsMask {test ip mask} {
      if {[string match "127.0.0.*" $test]} {
         return 1
      } ;# end if
      foreach {t} [split $test .] {a} [split $ip .] {A} [split $mask .] {
        if {($t&$A) != ($a&$A)} {
          return 0
        }
      }
      return 1
   }
   ;##
   ;# Determine if a url matches a pattern
   ;#
   proc localUrl {url} {
      regexp {(http://)*(.+)\?.*} $url -> -> url
      foreach patt [split $::snackAmpSettings(localurl) "\; "] {
         if {[string match -nocase $patt $url]} {
            return 1
         }
      } ;# end foreach
      return 0
   }
   variable myIP [myIP]
   variable defMask [subNetMask $myIP]
}

proc startlogerrors {args} {
   trace vdelete ::errorInfo w showlogerrors
   trace variable ::errorInfo w showlogerrors
}
proc showlogerrors {var index operand} {
   upvar $var ER
   puts "Error:$ER"
}
#startlogerrors
package provide gpUtils 1.0

