#==============================================================================
# Contains the implementation of the tablelist move and movecolumn subcommands.
#
# Copyright (c) 2003  Csaba Nemethi (E-mail: csaba.nemethi@t-online.de)
#==============================================================================

#------------------------------------------------------------------------------
# tablelist::moveSubCmd
#
# This procedure is invoked to process the tablelist move subcommand.
#------------------------------------------------------------------------------
proc tablelist::moveSubCmd {win source target} {
    upvar ::tablelist::ns${win}::data data

    if {$data(isDisabled) || $data(itemCount) == 0} {
	return ""
    }

    #
    # Adjust the indices to fit within the existing items and check them
    #
    if {$source > $data(lastRow)} {
	set source $data(lastRow)
    } elseif {$source < 0} {
	set source 0
    }
    if {$target > $data(itemCount)} {
	set target $data(itemCount)
    } elseif {$target < 0} {
	set target 0
    }
    if {$target == $source} {
	return -code error \
	       "cannot move item with index \"$source\" before itself"
    } elseif {$target == $source + 1} {
	return ""
    }

    #
    # Save some data of the entry widget if present in the source row
    #
    if {[set editRow $data(editRow)] == $source} {
	set editCol $data(editCol)
	saveEntryData $win
    }

    #
    # Check whether the source line is selected and then delete it
    #
    set w $data(body)
    set textIdx1 [expr {double($source + 1)}]
    set textIdx2 [expr {double($source + 2)}]
    if {[lsearch -exact [$w tag names $textIdx1] select] >= 0} {
	set selected 1
    } else {
	set selected 0
    }
    $w delete $textIdx1 $textIdx2

    #
    # Insert the source item before the target one
    #
    set target1 $target
    if {$source < $target} {
	incr target1 -1
    }
    set targetLine [expr {$target1 + 1}]
    $w insert $targetLine.0 \n
    set snipStr $data(-snipstring)
    set sourceItem [lindex $data(itemList) $source]
    set dispItem [strToDispStr $sourceItem]
    set key [lindex $sourceItem end]
    array set tagData [array get data $key*-\[bf\]*]		;# for speed
    set rowTags [array names tagData $key-\[bf\]*]
    set col 0
    foreach text [lrange $dispItem 0 $data(lastCol)] \
	    colFont $data(colFontList) \
	    colTags $data(colTagsList) \
	    fmtCmdFlag $data(fmtCmdFlagList) \
	    {pixels alignment} $data(colList) {
	if {$data($col-hide)} {
	    incr col
	    continue
	}

	#
	# Adjust the cell text and the image width
	#
	if {$fmtCmdFlag} {
	    set text [uplevel #0 $data($col-formatcommand) \
		      [list [lindex $sourceItem $col]]]
	    set text [strToDispStr $text]
	}
	if {[info exists data($key-$col-image)]} {
	    set image $data($key-$col-image)
	    set imageWidth [image width $image]
	} else {
	    set image ""
	    set imageWidth 0
	}
	if {[info exists data($key-$col-font)]} {
	    set cellFont $data($key-$col-font)
	} elseif {[info exists data($key-font)]} {
	    set cellFont $data($key-font)
	} else {
	    set cellFont $colFont
	}
	if {$pixels != 0} {			;# convention: static width
	    incr pixels $data($col-delta)
	}
	adjustElem $win text imageWidth $cellFont \
		   $pixels $alignment $snipStr

	#
	# Insert the text and the image
	#
	set cellTags [array names tagData $key-$col-\[bf\]*]
	set tagNames [concat $colTags $rowTags $cellTags]
	if {$imageWidth == 0} {
	    $w insert $targetLine.end \t$text\t $tagNames
	} else {
	    $w insert $targetLine.end \t\t $tagNames
	    insertElem $w $targetLine.end-1c $text $image $imageWidth $alignment
	}

	incr col
    }

    #
    # Update the item list
    #
    set data(itemList) [lreplace $data(itemList) $source $source]
    if {$target == $data(itemCount)} {
	lappend data(itemList) $sourceItem	;# this works much faster
    } else {
	set data(itemList) [linsert $data(itemList) $target1 $sourceItem]
    }

    #
    # Update the list variable if present
    #
    if {$data(hasListVar)} {
	trace vdelete ::$data(-listvariable) wu $data(listVarTraceCmd)
	upvar #0 $data(-listvariable) var
	set var [lreplace $var $source $source]
	set pureSourceItem [lrange $sourceItem 0 $data(lastCol)]
	if {$target == $data(itemCount)} {
	    lappend var $pureSourceItem		;# this works much faster
	} else {
	    set var [linsert $var $target1 $pureSourceItem]
	}
	trace variable ::$data(-listvariable) wu $data(listVarTraceCmd)
    }

    #
    # Update anchorIdx and activeIdx if needed
    #
    if {$data(anchorIdx) == $source} {
	set data(anchorIdx) $target1
    }
    if {$data(activeIdx) == $source} {
	set data(activeIdx) $target1
    }

    #
    # Restore the stripes in the body text widget
    #
    makeStripesWhenIdle $win

    #
    # Select the source item if it was selected before
    #
    if {$selected} {
	selectionSubCmd $win set $target1 $target1 2
    }

    #
    # Restore the entry widget if it was present before
    #
    if {$editRow == $source} {
	editcellSubCmd $win $target1 $editCol 1
    }

    return ""
}

#------------------------------------------------------------------------------
# tablelist::movecolumnSubCmd
#
# This procedure is invoked to process the tablelist movecolumn subcommand.
#------------------------------------------------------------------------------
proc tablelist::movecolumnSubCmd {win source target} {
    upvar ::tablelist::ns${win}::data data

    if {$data(isDisabled)} {
	return ""
    }

    #
    # Check the indices
    #
    if {$target == $source} {
	return -code error \
	       "cannot move column with index \"$source\" before itself"
    } elseif {$target == $source + 1} {
	return ""
    }

    #
    # Update the column list
    #
    set source3 [expr {3*$source}]
    set source3Plus2 [expr {$source3 + 2}]
    set target1 $target
    set target3 [expr {3*$target}]
    if {$source < $target} {
	incr target1 -1
	incr target3 -3
    }
    set sourceRange [lrange $data(-columns) $source3 $source3Plus2]
    set data(-columns) [lreplace $data(-columns) $source3 $source3Plus2]
    set data(-columns) [eval linsert {$data(-columns)} $target3 $sourceRange]

    #
    # Save the elements of data corresponding to source
    #
    array set tmp [array get data $source-*]
    array set tmp [array get data k*-$source-*]

    #
    # Build two lists of column numbers, neeeded
    # for shifting some elements of the data array
    #
    if {$source < $target} {
	for {set n $source} {$n < $target1} {incr n} {
	    lappend oldCols [expr {$n + 1}]
	    lappend newCols $n
	}
    } else {
	for {set n $source} {$n > $target} {incr n -1} {
	    lappend oldCols [expr {$n - 1}]
	    lappend newCols $n
	}
    }

    #
    # Set the elements of data corresponding to the columns in newCols to the
    # values corresponding to the columns with the same indices in oldCols
    #
    set w $data(body)
    foreach oldCol $oldCols newCol $newCols {
	foreach newName [array names data $newCol-*] {
	    unset data($newName)
	    $w tag delete $newName
	}
	foreach c [winfo children $data(hdrTxtFrLbl)$newCol] {
	    destroy $c
	}
	set data(fmtCmdFlagList) \
	    [lreplace $data(fmtCmdFlagList) $newCol $newCol 0]
	if {$data(editCol) == $oldCol} {
	    set data(editCol) $newCol
	}
	if {$data(arrowCol) == $oldCol} {
	    set data(arrowCol) $newCol
	}
	if {$data(sortCol) == $oldCol} {
	    set data(sortCol) $newCol
	}
	foreach oldName [array names data $oldCol-*] {
	    regsub $oldCol- $oldName $newCol- newName
	    set data($newName) $data($oldName)

	    set tail [lindex [split $oldName -] 1]
	    switch $tail {
		background -
		foreground {
		    $w tag configure $newName -$tail $data($newName)
		    $w tag raise $newName stripe
		}
		font {
		    $w tag configure $newName -$tail $data($newName)
		    $w tag lower $newName
		}
		formatcommand {
		    set data(fmtCmdFlagList) \
			[lreplace $data(fmtCmdFlagList) $newCol $newCol 1]
		}
		labelimage {
		    set tmp2($newCol-$tail) $data($newName)
		    unset data($newName)
		}
		selectbackground -
		selectforeground {
		    set tail [string range $tail 6 end]  ;# remove the select
		    $w tag configure $newName -$tail $data($newName)
		    $w tag raise $newName select
		}
	    }
	}
	foreach newName [array names data k*-$newCol-*] {
	    unset data($newName)
	    $w tag delete $newName
	}
	foreach oldName [array names data k*-$oldCol-*] {
	    regsub -- -$oldCol- $oldName -$newCol- newName
	    set data($newName) $data($oldName)

	    set tail [lindex [split $oldName -] 2]
	    switch $tail {
		background -
		foreground {
		    $w tag configure $newName -$tail $data($newName)
		    $w tag lower $newName disabled
		}
		font {
		    $w tag configure $newName -$tail $data($newName)
		    $w tag raise $newName
		}
		selectbackground -
		selectforeground {
		    set tail [string range $tail 6 end]  ;# remove the select
		    $w tag configure $newName -$tail $data($newName)
		    $w tag lower $newName disabled
		}
	    }
	}
    }

    #
    # Set the elements of data corresponding to
    # target1 to the values corresponding to source
    #
    foreach name [array names data $target1-*] {
	unset data($name)
	$w tag delete $name
    }
    foreach c [winfo children $data(hdrTxtFrLbl)$target1] {
	destroy $c
    }
    set data(fmtCmdFlagList) \
	[lreplace $data(fmtCmdFlagList) $target1 $target1 0]
    if {$data(editCol) == $source} {
	set data(editCol) $target1
    }
    if {$data(arrowCol) == $source} {
	set data(arrowCol) $target1
    }
    if {$data(sortCol) == $source} {
	set data(sortCol) $target1
    }
    foreach oldName [array names tmp $source-*] {
	regsub $source- $oldName $target1- newName
	set data($newName) $tmp($oldName)

	set tail [lindex [split $oldName -] 1]
	switch $tail {
	    background -
	    foreground {
		$w tag configure $newName -$tail $data($newName)
		$w tag raise $newName stripe
	    }
	    font {
		$w tag configure $newName -$tail $data($newName)
		$w tag lower $newName
	    }
	    formatcommand {
		set data(fmtCmdFlagList) \
		    [lreplace $data(fmtCmdFlagList) $target1 $target1 1]
	    }
	    labelimage {
		set tmp2($target1-$tail) $data($newName)
		unset data($newName)
	    }
	    selectbackground -
	    selectforeground {
		set tail [string range $tail 6 end]	;# remove the select
		$w tag configure $newName -$tail $data($newName)
		$w tag raise $newName select
	    }
	}
    }
    foreach name [array names data k*-$target1-*] {
	unset data($name)
	$w tag delete $name
    }
    foreach oldName [array names tmp k*-$source-*] {
	regsub -- -$source- $oldName -$target1- newName
	set data($newName) $tmp($oldName)

	set tail [lindex [split $oldName -] 2]
	switch $tail {
	    background -
	    foreground {
		$w tag configure $newName -$tail $data($newName)
		$w tag lower $newName disabled
	    }
	    font {
		$w tag configure $newName -$tail $data($newName)
		$w tag raise $newName
	    }
	    selectbackground -
	    selectforeground {
		set tail [string range $tail 6 end]	;# remove the select
		$w tag configure $newName -$tail $data($newName)
		$w tag lower $newName disabled
	    }
	}
    }

    #
    # Rebuild the lists of the column fonts and tag names
    #
    makeColFontAndTagLists $win

    #
    # Update the item list
    #
    foreach item $data(itemList) {
	set sourceText [lindex $item $source]
	set item [lreplace $item $source $source]
	set item [linsert $item $target1 $sourceText]
	lappend newItemList $item
    }
    set data(itemList) $newItemList

    #
    # Update the list variable if present
    #
    if {$data(hasListVar)} {
	trace vdelete ::$data(-listvariable) wu $data(listVarTraceCmd)
	upvar #0 $data(-listvariable) var
	set var {}
	foreach item $data(itemList) {
	    lappend var [lrange $item 0 $data(lastCol)]
	}
	trace variable ::$data(-listvariable) wu $data(listVarTraceCmd)
    }

    #
    # Set up and adjust the columns
    #
    setupColumns $win $data(-columns) 0
    adjustColumns $win {} 0

    #
    # Reconfigure the relevant column labels
    #
    foreach col [lappend newCols $target1] {
	foreach opt {-labelalign -labelbackground -labelborderwidth -labelfont
		     -labelforeground -labelheight -labelpady -labelrelief} {
	    if {[info exists data($col$opt)]} {
		doColConfig $col $win $opt $data($col$opt)
	    } else {
		doColConfig $col $win $opt ""
	    }
	}
	if {[info exists tmp2($col-labelimage)]} {
	    doColConfig $col $win -labelimage $tmp2($col-labelimage)
	}
    }

    #
    # Make sure the items will be redisplayed at idle time
    #
    redisplayWhenIdle $win

    return ""
}
